/*----------------------------------------------------------------------
  SDL2 Benchmark - Roberto Prieto 
  Copyright (C) 2013-2021 MegaStorm Systems
  contact@megastormsystems.com - http://www.megastormsystems.com

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

 ------------------------------------------------------------------------
                                SDL2.0.15 Benchmark Info
    
    - Uses SDL 2.0.15 (4/Feb/2021) using batching
    - With Windows 8 or greater OS, it also checks Direct3D11+triple-buffer
    - 64bits versions

    Changelog:
 
 v1.00 Feb/2021 - First version (using batching!)
   
 ------------------------------------------------------------------------ */

#include "SDL/SDL.h"
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <windows.h>
#include <VersionHelpers.h>
#include "../SDL-Benchmark.h"

// Common vars
static SDL_Rect *positions;
static SDL_Rect *velocities;
static int sprite_w, sprite_h;
SDLBenchParameters SDLBenchParams; // Parameters struct

// SDL2 vars
SDL_Window *window;
SDL_Renderer *renderer;
SDL_Texture **pSpriteTable = NULL;
//#define DEBUG_RENDERING 

SDL_bool Init(const char *renderdriver)
{
    int j, m, n;

    if(SDL_VideoInit(NULL) < 0) 
    {
        printf("Couldn't initialize video driver: %s\n",SDL_GetError());
        return SDL_FALSE;
    }
    window = SDL_CreateWindow("SDL 2.0.15 Benchmarks", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, SDLBenchParams.iVideoWidth, SDLBenchParams.iVideoHeight, 0); // SDL_WINDOW_FULLSCREEN
    if(!window) 
    {
        printf("Couldn't create window: %s\n", SDL_GetError());
        return SDL_FALSE;
    }
    SDL_ShowWindow(window);
    renderer = NULL;

    m = -1;
    if(renderdriver) 
    {
        SDL_RendererInfo info;
        n = SDL_GetNumRenderDrivers();
        for(j = 0; j < n; ++j) 
        {
            SDL_GetRenderDriverInfo(j, &info);
            if(SDL_strcasecmp(info.name, renderdriver) == 0) 
            {
                m = j;
                break;
            }
        }
        if(m == n) 
        {
            printf("Couldn't find render driver named %s", renderdriver);
            return SDL_FALSE;
        }
    }
    SDL_SetHint(SDL_HINT_RENDER_BATCHING, "1");
    renderer = NULL;
    renderer = SDL_CreateRenderer(window, m, 0); // No vsync
    if(!renderer) 
    {
        printf("Couldn't create renderer: %s\n", SDL_GetError());
        return SDL_FALSE;
    }
    SDL_ShowCursor(0);
    return SDL_TRUE;
}

int RemoveSprites(int iNumRealSpr)
{
    int i = 0;

    while(i < iNumRealSpr)
    {
        if(pSpriteTable != NULL)
        {
            if(pSpriteTable[i] != NULL) SDL_DestroyTexture(pSpriteTable[i]);
            pSpriteTable[i] = NULL;
        }
        ++i;        
    }
    free(pSpriteTable);    
    pSpriteTable = NULL;
    return 0;
}

int LoadSprites(char *file, int iNumRealSpr)
{
    SDL_Surface *temp;
    int i = 0;

    // Load the sprite image
    temp = SDL_LoadBMP(file);
    if(temp == NULL) 
    {
        fprintf(stderr, "Couldn't load %s: %s", file, SDL_GetError());
        return (-1);
    }
    sprite_w = temp->w;
    sprite_h = temp->h;

    // Set transparent pixel as the pixel at (0,0)
    if(temp->format->palette) 
    {
        SDL_SetColorKey(temp, 1, *(Uint8 *) temp->pixels);
    } 
    else 
    {
        switch(temp->format->BitsPerPixel) 
        {
        case 15:
            SDL_SetColorKey(temp, 1, (*(Uint16 *) temp->pixels) & 0x00007FFF);
            break;
        case 16:
            SDL_SetColorKey(temp, 1, *(Uint16 *) temp->pixels);
            break;
        case 24:
            SDL_SetColorKey(temp, 1, (*(Uint32 *) temp->pixels) & 0x00FFFFFF);
            break;
        case 32:
            SDL_SetColorKey(temp, 1, *(Uint32 *) temp->pixels);
            break;
        }
    }

    // Create textures for each sprite using the surface
    int bFlagError = 0;
    pSpriteTable = new SDL_Texture *[iNumRealSpr];    
    while(i < iNumRealSpr)
    {
        pSpriteTable[i] = SDL_CreateTextureFromSurface(renderer, temp);
        if(!pSpriteTable[i]) 
        {            
            fprintf(stderr, "Couldn't create texture for sprite %d: %s\n", i, SDL_GetError());
            bFlagError = 1;
        }
        ++i;        
    }
    SDL_FreeSurface(temp);
    if(bFlagError == 1) 
    {
        RemoveSprites(iNumRealSpr);
        return -1;
    }
    return 0;
}

void MoveSprites(int iNumSpr, int iNumRealSpr)
{
    int i;
    SDL_Rect *position, *velocity;
    #ifdef DEBUG_RENDERING
     LARGE_INTEGER StartingTime, EndingTime, measure1, measure2;
     QueryPerformanceCounter(&StartingTime);
    #endif

    // Grey background    
    SDL_SetRenderDrawColor(renderer, 0xA0, 0xA0, 0xA0, 0xFF);
    SDL_RenderClear(renderer);
    #ifdef DEBUG_RENDERING
     QueryPerformanceCounter(&measure1);
    #endif

    // Move the sprite, bounce at the wall, and draw
    for(i = 0; i < iNumSpr; ++i) 
    {
        position = &positions[i];
        velocity = &velocities[i];
        position->x += velocity->x;
        if((position->x < 0) || (position->x >= (SDLBenchParams.iVideoWidth - sprite_w))) 
        {
            velocity->x = -velocity->x;
            position->x += velocity->x;
        }
        position->y += velocity->y;
        if((position->y < 0) || (position->y >= (SDLBenchParams.iVideoHeight - sprite_h))) 
        {
            velocity->y = -velocity->y;
            position->y += velocity->y;
        }

        // Blit the sprite onto the screen
        if(iNumRealSpr == 1) SDL_RenderCopy(renderer, pSpriteTable[0], NULL, position);
        else SDL_RenderCopy(renderer, pSpriteTable[i], NULL, position);
    }
    #ifdef DEBUG_RENDERING
     QueryPerformanceCounter(&measure2);
    #endif

    // Update the screen!
    SDL_RenderPresent(renderer);
    
    #ifdef DEBUG_RENDERING
     QueryPerformanceCounter(&EndingTime);
     printf("FrameTime: %llu + %llu + %llu = %llu\n", 
        measure1.QuadPart - StartingTime.QuadPart, 
        measure2.QuadPart - measure1.QuadPart, 
        EndingTime.QuadPart - measure2.QuadPart, 
        EndingTime.QuadPart - StartingTime.QuadPart);     
    #endif
}

int main(int argc, char *argv[])
{
    int i, done, iMaxSprites, iRealSprites;
    Uint32 iStart, iFrames;
    double fFPS_1, fFPS_2;
    SDL_RendererInfo ren;
    int iRenderer = 0;
    char *renderdriver;    

    // Print information and argument parsing
    #ifdef _WIN64
      parseAndInformation(argc, argv, SDLBENCH_SDL2B_VERSION" 64bits Benchmark", &SDLBenchParams);
    #else
      parseAndInformation(argc, argv, SDLBENCH_SDL2B_VERSION" 32bits Benchmark", &SDLBenchParams);
    #endif
    
    // Max and real sprites
    iMaxSprites = SDLBenchParams.iNumSprites;
    if(iMaxSprites < SDLBenchParams.iNumSprites2) iMaxSprites = SDLBenchParams.iNumSprites2;
    if(SDLBenchParams.iFillRateMode == 0) iRealSprites = iMaxSprites;
    else iRealSprites = 1;

    // First test D3D
    renderdriver = "direct3d";    

REINIT_SDL:
    // Initialize 
    if(!Init(renderdriver)) return -1;
    SDL_GetRendererInfo(renderer, &ren);
    printf("SDL2 renderer: %s\n", ren.name);
    SDL_SetRenderDrawColor(renderer, 0xA0, 0xA0, 0xA0, 0xFF);
    SDL_RenderClear(renderer);
    if(LoadSprites(SDLBenchParams.szIcon, iRealSprites) < 0) 
    {
        SDL_DestroyRenderer(renderer);
        SDL_DestroyWindow(window);
        SDL_VideoQuit();
        return -2;
    }

    // Allocate memory for the sprite info and initialize it    
    positions = (SDL_Rect *) SDL_malloc(iMaxSprites * sizeof(SDL_Rect));
    velocities = (SDL_Rect *) SDL_malloc(iMaxSprites * sizeof(SDL_Rect));
    
    srand((unsigned int)time(NULL));
    for(i = 0; i < iMaxSprites; ++i) 
    {
        positions[i].x = rand() % (SDLBenchParams.iVideoWidth - sprite_w);
        positions[i].y = rand() % (SDLBenchParams.iVideoHeight - sprite_h);
        positions[i].w = sprite_w;
        positions[i].h = sprite_h;
        velocities[i].x = 0;
        velocities[i].y = 0;
        while(!velocities[i].x && !velocities[i].y) 
        {
            velocities[i].x = (rand() % (MAX_SPEED * 2 + 1)) - MAX_SPEED;
            velocities[i].y = (rand() % (MAX_SPEED * 2 + 1)) - MAX_SPEED;
        }
    }
    
    // ---Main render loop for iNumSprites---    
    iFrames = 0;	
    iStart = clock();	
    done = 0;	
    while(!done) 
    {				
        ++iFrames;          
        MoveSprites(SDLBenchParams.iNumSprites, iRealSprites);
        if(( (float)(clock() - iStart) / 1000) >= SDLBenchParams.dBaseTime) { done = 1; }        
    }
    // Print out some timing information
    if(clock() > iStart) 
    {
        fFPS_1 = (double)iFrames / (double)((clock() - iStart) / 1000.0);
        printf(" - Moving %d sprites at %2.2f frames per second\n", SDLBenchParams.iNumSprites, fFPS_1);
    }	    
    // ------

    // ---Main render loop for iNumSprites2---
    iFrames = 0;	
    iStart = clock();	
    done = 0;
    while(!done) 
    {				
        ++iFrames;
        MoveSprites(SDLBenchParams.iNumSprites2, iRealSprites);
        if(( (float)(clock() - iStart) / 1000) >= SDLBenchParams.dBaseTime) { done = 1; }       
    }
    // Print out some timing information
    if(clock() > iStart) 
    {
        fFPS_2 = (double)iFrames / (double)((clock() - iStart) / 1000.0);
        printf(" - Moving %d sprites at %2.2f frames per second\n\n", SDLBenchParams.iNumSprites2, fFPS_2);
    }
    // ------

    // Integrated mode: extended information output
    if(SDLBenchParams.iIntegrated)
    {
        #ifdef _WIN64
          // Print information and argument parsing
          FILE *fp = fopen("sdl2-64.table","at");
          fprintf(fp,"          <tr><td align=center><b><font size=2 color=#000080>%s 64bits - %s", SDLBENCH_SDL2B_VERSION, ren.name);
        #else
          // Print information and argument parsing
          FILE *fp = fopen("sdl2.table","at");
          fprintf(fp,"          <tr><td align=center><b><font size=2 color=#000080>%s - %s", SDLBENCH_SDL2015_VERSION, ren.name);
        #endif        
        fprintf(fp,"</font></b></td>");
        fprintf(fp,"<td align=center><font size=2 color=#800000>%2.2f</font></td>",fFPS_1);
        fprintf(fp,"<td align=center><font size=2 color=#800000>%2.2f</font></td></tr>\n",fFPS_2);
        fclose(fp);
        fp = fopen("SDL-Benchmark.graph","at");
        #ifdef _WIN64
          fprintf(fp,"%s 64bits (%s), %2.2f, %2.2f\n", SDLBENCH_SDL2B_VERSION, ren.name, fFPS_1, fFPS_2);
        #else
          fprintf(fp,"%s (%s), %2.2f, %2.2f\n", SDLBENCH_SDL2015_VERSION, ren.name, fFPS_1, fFPS_2);
        #endif        
        fclose(fp);
    }
    
    // Close and free memory buffers
    SDL_free(positions);
    SDL_free(velocities);
    RemoveSprites(iRealSprites);
    SDL_DestroyRenderer(renderer);
    SDL_DestroyWindow(window);
    SDL_VideoQuit();

    // Renderer selector
    switch(iRenderer)
    {
    case 0:
        ++iRenderer; // Next renderer
        renderdriver = "direct3d11";
        if (IsWindows8OrGreater())
        {
            printf("Windows 8+ detected, running Direct3D11 test...\n");
            goto REINIT_SDL;
        }
    case 1: 
        ++iRenderer; // Next renderer
        renderdriver = "opengl";
        goto REINIT_SDL;
        break;
    }   

    // End!     
    return 0;
}

